//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/ParameterTreeUtils.cpp
//! @brief     Implements ParameterTreeUtils namespace
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/ParameterTreeUtils.h"
#include "GUI/Model/Fit/ParameterTreeItems.h"
#include "GUI/Model/Group/GroupItem.h"
#include "GUI/Model/Group/PropertyItem.h"
#include "GUI/Model/Instrument/InstrumentItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Sample/FTDistributionItems.h"
#include "GUI/Model/Sample/InterferenceItemCatalog.h"
#include "GUI/Model/Sample/InterferenceItems.h"
#include "GUI/Model/Sample/Lattice2DItems.h"
#include "GUI/Model/Sample/LayerItem.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Sample/ParticleLayoutItem.h"
#include "GUI/Model/Types/VectorDescriptor.h"
#include "GUI/Util/Error.h"
#include <QStack>
#include <boost/polymorphic_cast.hpp>

using boost::polymorphic_downcast;

namespace {

QString labelWithUnit(const QString& label, variant<QString, Unit> unit)
{
    const QString s = std::holds_alternative<QString>(unit) ? std::get<QString>(unit)
                                                            : unitAsString(std::get<Unit>(unit));

    if (!s.isEmpty())
        return label + "[" + s + "]";

    return label;
}

void handleItem(ParameterContainerItem* container, QObject* tree, SessionItem* source,
                bool recreateBackupValues)
{
    if (auto* p = dynamic_cast<ParameterLabelItem*>(tree))
        p->setTitle(source->itemName());

    else if (auto* parItem = dynamic_cast<ParameterItem*>(tree)) {
        parItem->setTitle(source->itemName());

        const double sourceValue = source->value().toDouble();
        parItem->linkToSessionItem(source);
        if (recreateBackupValues)
            container->setBackupValue(parItem->link(), sourceValue);
        return;
    }

    else
        return;

    for (SessionItem* child : source->children()) {
        if (child->isVisible() && child->isEnabled()) {
            if (child->hasModelType<PropertyItem>()) {
                if (child->value().type() == QVariant::Double) {
                    auto* p = new ParameterItem(tree);
                    handleItem(container, p, child, recreateBackupValues);
                }
            } else if (child->hasModelType<GroupItem>()) {
                SessionItem* currentItem = dynamic_cast<GroupItem*>(child)->currentItem();
                if (currentItem && currentItem->numberOfChildren() > 0) {
                    auto* p = new ParameterLabelItem(tree);
                    handleItem(container, p, currentItem, recreateBackupValues);
                }
            } else {
                auto* p = new ParameterLabelItem(tree);
                handleItem(container, p, child, recreateBackupValues);
            }
        }
    }
}

} // namespace

ParameterTreeBuilder::ParameterTreeBuilder(JobItem* jobItem, bool recreateBackupValues)
    : m_jobItem(jobItem), m_recreateBackupValues(recreateBackupValues)
{
}

void ParameterTreeBuilder::build()
{
    addMaterials();
    addSample();
    addInstrument();
}

void ParameterTreeBuilder::addMaterials()
{
    auto* materialTopLabel =
        new ParameterLabelItem("Materials", parameterContainer()->parameterTreeRoot());
    for (auto* material : m_jobItem->materialItems().materialItems()) {
        auto* materialLabel = new ParameterLabelItem(material->materialName(), materialTopLabel);
        if (material->hasRefractiveIndex()) {
            addParameterItem(materialLabel, material->delta());
            addParameterItem(materialLabel, material->beta());
        } else {
            addParameterItem(materialLabel, material->sldRe());
            addParameterItem(materialLabel, material->sldIm());
        }

        if (allowMagneticFields())
            addParameterItem(materialLabel, material->magnetizationVector());
    }
}

void ParameterTreeBuilder::addSample()
{
    auto* sampleTopLabel =
        new ParameterLabelItem("Sample", parameterContainer()->parameterTreeRoot());
    addParameterItem(sampleTopLabel, m_jobItem->sampleItem()->crossCorrLength());
    if (allowMagneticFields())
        addParameterItem(sampleTopLabel, m_jobItem->sampleItem()->externalFieldVector());

    int iLayer = 0;
    for (auto* layer : m_jobItem->sampleItem()->layers()) {
        auto* layerLabel = new ParameterLabelItem("Layer" + QString::number(iLayer++),
                                                  parameterContainer()->parameterTreeRoot());
        if (!layer->isTopLayer() && !layer->isBottomLayer())
            addParameterItem(layerLabel, layer->thickness());
        if (!layer->isTopLayer())
            if (auto* roughnessItem = layer->roughness().currentItem()) {
                auto* roughnessLabel = new ParameterLabelItem("Top roughness", layerLabel);
                addParameterItem(roughnessLabel, roughnessItem->sigma());
                addParameterItem(roughnessLabel, roughnessItem->hurst());
                addParameterItem(roughnessLabel, roughnessItem->lateralCorrelationLength());
            }

        int iLayout = 0;
        for (auto* layout : layer->layouts()) {
            auto* layoutLabel =
                new ParameterLabelItem("Layout" + QString::number(iLayout++), layerLabel);
            if (!layout->totalDensityIsDefinedByInterference())
                addParameterItem(layoutLabel, layout->ownDensity());
            addParameterItem(layoutLabel, layout->weight());

            addInterference(layoutLabel, layout);

            for (auto* p : layout->particles())
                addParticle(layoutLabel, p);
        }
    }
}

void ParameterTreeBuilder::addInstrument()
{
    auto* sourceLabel = new ParameterLabelItem(parameterContainer()->parameterTreeRoot());
    handleItem(parameterContainer(), sourceLabel, m_jobItem->instrumentItem(),
               m_recreateBackupValues);
}

void ParameterTreeBuilder::addParameterItem(ParameterLabelItem* parent, DoubleDescriptor d)
{
    auto* parameterItem = new ParameterItem(parent);
    parameterItem->setTitle(labelWithUnit(d.label, d.unit));
    parameterItem->linkToDescriptor(d);
    if (m_recreateBackupValues)
        m_jobItem->parameterContainerItem()->setBackupValue(parameterItem->link(), d.get());
}

void ParameterTreeBuilder::addParameterItem(ParameterLabelItem* parent, VectorDescriptor d)
{
    addParameterItem(parent, d.x);
    addParameterItem(parent, d.y);
    addParameterItem(parent, d.z);
}

ParameterContainerItem* ParameterTreeBuilder::parameterContainer()
{
    return m_jobItem->parameterContainerItem();
}

bool ParameterTreeBuilder::allowMagneticFields() const
{
    // TODO: return true when specular instrument is ready for magnetization
    return !m_jobItem->isSpecularJob();
}

void ParameterTreeBuilder::addInterference(ParameterLabelItem* layoutLabel,
                                           const ParticleLayoutItem* layout)
{
    if (!layout->interference().currentItem())
        return;

    const auto itfType = InterferenceItemCatalog::type(layout->interference().currentItem());
    const QString title = InterferenceItemCatalog::uiInfo(itfType).menuEntry;

    auto* itfLabel = new ParameterLabelItem(title, layoutLabel);

    Q_UNUSED(itfLabel); // temporary
    // later in handleItems the rotation of an interference may have to be ignored.
    // #baMigration implement the following!
    //         if (auto* c = dynamic_cast<Interference2DAbstractLatticeItem*>(child))
    //             if (c->latticeType().currentItem())
    //                 c->latticeType().currentItem()->enableRotationAngle(!c->xiIntegration());

    // #baMigration see the following!
    // if PDF1 and PDF2 are of same type, add a number to the name
    // if (auto* itf = dynamic_cast<Interference2DParaCrystalItem*>(
    //        layout->interference().currentItem())) {
    //    SessionItem* pdf1 = itf->probabilityDistribution1().currentItem();
    //    SessionItem* pdf2 = itf->probabilityDistribution2().currentItem();
    //
    //    if (pdf1 && pdf2) {
    //        if (pdf1->modelType() == pdf2->modelType()) {
    //            pdf1->setDisplayName(pdf1->modelType() + "0");
    //            pdf2->setDisplayName(pdf2->modelType() + "1");
    //        } else {
    //            pdf1->setDisplayName(pdf1->modelType());
    //            pdf2->setDisplayName(pdf2->modelType());
    //        }
    //    }
    //}

    throw std::logic_error("The method or operation is not implemented.");
}

void ParameterTreeBuilder::addParticle(ParameterLabelItem* parentLabel, ItemWithParticles* p)
{
    Q_UNUSED(parentLabel); // temporary
    Q_UNUSED(p);           // temporary
    // #baMigration implement the following!
    // if (auto* c = dynamic_cast<ItemWithParticles*>(child))
    //     c->enableAbundance(!c->parentHasOwnAbundance());
    //
    // function was:
    //
    //         bool ItemWithParticles::parentHasOwnAbundance() const
    //         {
    //             return dynamic_cast<ParticleCoreShellItem*>(parent())
    //                    || dynamic_cast<ParticleCompositionItem*>(parent())
    //                    || dynamic_cast<MesoCrystalItem*>(parent());
    //         }


    //
    // later in handleItems the position of a shell particle has to be ignored.
    // #baMigration implement the following!
    //         if (auto* c = dynamic_cast<ParticleCoreShellItem*>(child))
    //             if (c->shell())
    //                 c->shell()->positionItem()->setEnabled(false);

    // later in handleItems the rotation of an interference may have to be ignored.
    // #baMigration implement the following!
    //         if (auto* c = dynamic_cast<Interference2DAbstractLatticeItem*>(child))
    //             if (c->latticeType().currentItem())
    //                 c->latticeType().currentItem()->enableRotationAngle(!c->xiIntegration());


    throw std::logic_error("The method or operation is not implemented.");
}
