//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/FTDecayFunctionItemCatalogs.cpp
//! @brief     Implements FTDecayFunctionItemCatalog classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/FTDecayFunctionItemCatalogs.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Sample/FTDecayFunctionItems.h"

FTDecayFunction1DItemCatalog::CatalogedType* FTDecayFunction1DItemCatalog::create(Type type)
{
    switch (type) {
    case Type::Cauchy:
        return new FTDecayFunction1DCauchyItem();
    case Type::Gauss:
        return new FTDecayFunction1DGaussItem();
    case Type::Triangle:
        return new FTDecayFunction1DTriangleItem();
    case Type::Voigt:
        return new FTDecayFunction1DVoigtItem();
    default:
        ASSERT(false);
    }
}

QVector<FTDecayFunction1DItemCatalog::Type> FTDecayFunction1DItemCatalog::types()
{
    return {Type::Cauchy, Type::Gauss, Type::Triangle, Type::Voigt};
}

FTDecayFunction1DItemCatalog::UiInfo FTDecayFunction1DItemCatalog::uiInfo(Type type)
{
    auto createUiInfo = [](const QString& menuEntry, const QString& description) {
        UiInfo info;
        info.menuEntry = menuEntry;
        info.description = description;
        return info;
    };

    switch (type) {
    case Type::Cauchy:
        return createUiInfo("Cauchy 1D", "One-dimensional Cauchy decay function");
    case Type::Gauss:
        return createUiInfo("Gauss 1D", "One-dimensional Gauss decay function");
    case Type::Triangle:
        return createUiInfo("Triangle 1D", "One-dimensional triangle decay function");
    case Type::Voigt:
        return createUiInfo("Voigt 1D", "One-dimensional pseudo-Voigt decay function");
    default:
        ASSERT(false);
    }
}

FTDecayFunction1DItemCatalog::Type FTDecayFunction1DItemCatalog::type(CatalogedType* item)
{
    ASSERT(item);

#define CHECK(type)                                                                                \
    if (dynamic_cast<FTDecayFunction1D##type##Item*>(item))                                        \
    return Type::type

    CHECK(Cauchy);
    CHECK(Gauss);
    CHECK(Triangle);
    CHECK(Voigt);
#undef CHECK

    ASSERT(false);
}

/* --------------------------------------------------------------------------------------------- */

FTDecayFunction2DItemCatalog::CatalogedType* FTDecayFunction2DItemCatalog::create(Type type)
{
    switch (type) {
    case Type::Cauchy:
        return new FTDecayFunction2DCauchyItem();
    case Type::Gauss:
        return new FTDecayFunction2DGaussItem();
    case Type::Voigt:
        return new FTDecayFunction2DVoigtItem();
    default:
        ASSERT(false);
    }
}

QVector<FTDecayFunction2DItemCatalog::Type> FTDecayFunction2DItemCatalog::types()
{
    return {Type::Cauchy, Type::Gauss, Type::Voigt};
}

FTDecayFunction2DItemCatalog::UiInfo FTDecayFunction2DItemCatalog::uiInfo(Type type)
{
    auto createUiInfo = [](const QString& menuEntry, const QString& description) {
        UiInfo info;
        info.menuEntry = menuEntry;
        info.description = description;
        return info;
    };

    switch (type) {
    case Type::Cauchy:
        return createUiInfo("Cauchy 2D", "Two-dimensional Cauchy decay function");
    case Type::Gauss:
        return createUiInfo("Gauss 2D", "Two-dimensional Gauss decay function");
    case Type::Voigt:
        return createUiInfo("Voigt 2D", "Two-dimensional pseudo-Voigt decay function");
    default:
        ASSERT(false);
    }
}

FTDecayFunction2DItemCatalog::Type FTDecayFunction2DItemCatalog::type(const CatalogedType* item)
{
    ASSERT(item);

    if (dynamic_cast<const FTDecayFunction2DCauchyItem*>(item))
        return Type::Cauchy;
    if (dynamic_cast<const FTDecayFunction2DGaussItem*>(item))
        return Type::Gauss;
    if (dynamic_cast<const FTDecayFunction2DVoigtItem*>(item))
        return Type::Voigt;

    ASSERT(false);
}
