//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/FTDistributionItemCatalogs.cpp
//! @brief     Implements FTDistributionItemCatalog classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/FTDistributionItemCatalogs.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Sample/FTDistributionItems.h"

FTDistribution1DItemCatalog::CatalogedType* FTDistribution1DItemCatalog::create(Type type)
{
    switch (type) {
    case Type::Cauchy:
        return new FTDistribution1DCauchyItem();
    case Type::Gauss:
        return new FTDistribution1DGaussItem();
    case Type::Gate:
        return new FTDistribution1DGateItem();
    case Type::Triangle:
        return new FTDistribution1DTriangleItem();
    case Type::Cosine:
        return new FTDistribution1DCosineItem();
    case Type::Voigt:
        return new FTDistribution1DVoigtItem();
    default:
        ASSERT(false);
    }
}

QVector<FTDistribution1DItemCatalog::Type> FTDistribution1DItemCatalog::types()
{
    return {Type::Cauchy, Type::Gauss, Type::Gate, Type::Triangle, Type::Cosine, Type::Voigt};
}

FTDistribution1DItemCatalog::UiInfo FTDistribution1DItemCatalog::uiInfo(Type type)
{
    auto createUiInfo = [](const QString& menuEntry, const QString& description) {
        UiInfo info;
        info.menuEntry = menuEntry;
        info.description = description;
        return info;
    };

    switch (type) {
    case Type::Cauchy:
        return createUiInfo("Cauchy 1D", "One-dimensional Cauchy probability distribution");
    case Type::Gauss:
        return createUiInfo("Gauss 1D", "One-dimensional Gauss probability distribution");
    case Type::Gate:
        return createUiInfo("Gate 1D", "One-dimensional Gate probability distribution");
    case Type::Triangle:
        return createUiInfo("Triangle 1D", "One-dimensional triangle probability distribution");
    case Type::Cosine:
        return createUiInfo("Cosine 1D", "One-dimensional Cosine probability distribution");
    case Type::Voigt:
        return createUiInfo("Voigt 1D", "One-dimensional pseudo-Voigt probability distribution");
    default:
        ASSERT(false);
    }
}

FTDistribution1DItemCatalog::Type FTDistribution1DItemCatalog::type(CatalogedType* item)
{
    ASSERT(item);

#define CHECK(type)                                                                                \
    if (dynamic_cast<FTDistribution1D##type##Item*>(item))                                         \
    return Type::type

    CHECK(Cauchy);
    CHECK(Gauss);
    CHECK(Gate);
    CHECK(Triangle);
    CHECK(Cosine);
    CHECK(Voigt);
#undef CHECK

    ASSERT(false);
}

/* --------------------------------------------------------------------------------------------- */

FTDistribution2DItemCatalog::CatalogedType* FTDistribution2DItemCatalog::create(Type type)
{
    switch (type) {
    case Type::Cauchy:
        return new FTDistribution2DCauchyItem();
    case Type::Gauss:
        return new FTDistribution2DGaussItem();
    case Type::Gate:
        return new FTDistribution2DGateItem();
    case Type::Cone:
        return new FTDistribution2DConeItem();
    case Type::Voigt:
        return new FTDistribution2DVoigtItem();
    default:
        ASSERT(false);
    }
}

QVector<FTDistribution2DItemCatalog::Type> FTDistribution2DItemCatalog::types()
{
    return {Type::Cauchy, Type::Gauss, Type::Gate, Type::Cone, Type::Voigt};
}

FTDistribution2DItemCatalog::UiInfo FTDistribution2DItemCatalog::uiInfo(Type type)
{
    auto createUiInfo = [](const QString& menuEntry, const QString& description) {
        UiInfo info;
        info.menuEntry = menuEntry;
        info.description = description;
        return info;
    };

    switch (type) {
    case Type::Cauchy:
        return createUiInfo("Cauchy 2D", "Two-dimensional Cauchy probability distribution");
    case Type::Gauss:
        return createUiInfo("Gauss 2D", "Two-dimensional Gauss probability distribution");
    case Type::Gate:
        return createUiInfo("Gate 2D", "Two-dimensional Gate probability distribution");
    case Type::Cone:
        return createUiInfo("Cone 2D", "Two-dimensional Cone probability distribution");
    case Type::Voigt:
        return createUiInfo("Voigt 2D", "Two-dimensional pseudo-Voigt probability distribution");
    default:
        ASSERT(false);
    }
}

FTDistribution2DItemCatalog::Type FTDistribution2DItemCatalog::type(const CatalogedType* item)
{
    ASSERT(item);

    if (dynamic_cast<const FTDistribution2DCauchyItem*>(item))
        return Type::Cauchy;
    if (dynamic_cast<const FTDistribution2DGaussItem*>(item))
        return Type::Gauss;
    if (dynamic_cast<const FTDistribution2DGateItem*>(item))
        return Type::Gate;
    if (dynamic_cast<const FTDistribution2DConeItem*>(item))
        return Type::Cone;
    if (dynamic_cast<const FTDistribution2DVoigtItem*>(item))
        return Type::Voigt;

    ASSERT(false);
}
